
/*
 * This contains roughly the same structure as the main decoder,
 * except that instead of executing the instruction, we write 
 * an assembler-like representation of it to a passed buffer.
 *
 */

#include "sim.h"


/* [row][col][stringlen] */
static char regs[16][4][15] = {
  {"ax0", "i0", "i4", "astat"},
  {"ax1", "i1", "i5", "mstat"},
  {"mx0", "i2", "i6", "sstat"},
  {"mx1", "i3", "i7", "imask"},
  {"ay0", "m0", "m4", "icntl"},
  {"ay1", "m1", "m5", "cntr"},
  {"my0", "m2", "m6", "sb"},
  {"my1", "m3", "m7", "px"},
  {"si", "l0", "l4", "rx0"},
  {"se", "l1", "l5", "tx0"},
  {"ar", "l2", "l6", "rx1"},
  {"mr0", "l3", "l7", "rx1"},
  {"mr1", "", "", "ifc"},
  {"mr2", "", "", "owrcntr"},
  {"sr0", "pmovlay", "", ""},
  {"sr1", "", "", ""}
};

static char xops[8][15] = {
  "x0",
  "x1",
  "ar",
  "mr0",
  "mr1",
  "mr2",
  "sr0",
  "sr1"
};

static char yops[4][15] = {
  "y0",
  "y1",
  "f",
  "zero"
};

static char sfs[16][3][15] = {
  {"lshift", "(hi)", {0}},
  {"lshift", "(hi)", {1}},
  {"lshift", "(lo)", {0}},
  {"lshift", "(lo)", {1}},
  {"ashift", "(hi)", {0}},
  {"ashift", "(hi)", {1}},
  {"ashift", "(lo)", {0}},
  {"ashift", "(lo)", {1}},
  {"norm", "(hi)", {0}},
  {"norm", "(hi)", {1}},
  {"norm", "(lo)", {0}},
  {"norm", "(lo)", {1}},
  {"exp", "(hi)", {0}},
  {"exp", "(hix)", {0}},
  {"exp", "(lo)", {0}},
  {"dbe", "", {0}}
};

static char conds[16][15] = {
  "eq",
  "ne", 
  "gt", 
  "le",
  "lt",
  "ge",
  "av",
  "not av",
  "ac", 
  "not ac",
  "neg",
  "pos",
  "mv",
  "not mv",
  "not ce",
  "" /* always true */
};

static char flagops[4][15] = {
  "nochange",
  "toggle",
  "reset",
  "set"
};

/*
 * Returns the reg/rgp for the register name (of len namelen).
 *
 */
int findreg(char *name, int namelen, int *reg, int *rgp)
{
  int i,j;
  if (!name || (!reg || !rgp))
    return 0;

  for (i=0;i<16;i++)
    {
      for (j=0;j<4;j++)
	{
	  if (!strncmp(name, regs[i][j], namelen))
	    {
	      *reg = i;
	      *rgp = j;
	      return 1;
	    }
	}
    }
  
  return 0;
}

unsigned short getconstant(unsigned int yy, unsigned int cc, unsigned int bo)
{
  if(bo==1)
    {
      switch(yy)
	{
	case 0x00:
	  switch(cc)
	    {
	    case 0x00: return 0x0001;
	    case 0x01: return 0x0002;
	    case 0x02: return 0x0004;
	    case 0x03: return 0x0008;
	    }
	case 0x01:
	  switch(cc)
	    {
	    case 0x00: return 0x0010;
	    case 0x01: return 0x0020;
	    case 0x02: return 0x0040;
	    case 0x03: return 0x0080;
	    }
	case 0x02:
	  switch(cc)
	    {
	    case 0x00: return 0x0100;
	    case 0x01: return 0x0200;
	    case 0x02: return 0x0400;
	    case 0x03: return 0x0800;
	    }
	case 0x03:
	  switch(cc)
	    {
	    case 0x00: return 0x1000;
	    case 0x01: return 0x2000;
	    case 0x02: return 0x4000;
	    case 0x03: return 0x8000;
	    }
	}
    }
  else if (bo==3)
    {
      switch(yy)
	{
	case 0x00:
	  switch(cc)
	    {
	    case 0x00: return 0xfffe;
	    case 0x01: return 0xfffd;
	    case 0x02: return 0xfffb;
	    case 0x03: return 0xfff7;
	    }
	case 0x01:
	  switch(cc)
	    {
	    case 0x00: return 0xffef;
	    case 0x01: return 0xffdf;
	    case 0x02: return 0xffbf;
	    case 0x03: return 0xff7f;
	    }
	case 0x02:
	  switch(cc)
	    {
	    case 0x00: return 0xfeff;
	    case 0x01: return 0xfdff;
	    case 0x02: return 0xfbff;
	    case 0x03: return 0xf7ff;
	    }
	case 0x03:
	  switch(cc)
	    {
	    case 0x00: return 0xefff;
	    case 0x01: return 0xdfff;
	    case 0x02: return 0xbfff;
	    case 0x03: return 0x7fff;
	    }
	}
    }
  else
    fprintf(stderr, "pretty: unknown bo\n");
  return 0;
}

int prettydecode(unsigned int opcode, char *buf)
{
  if (!buf)
    return -1;

  memset(buf, 0x00, MAXDECODE+1);

  /* No-oper codes... */
  if ((opcode >> 16) == 0x0000)
    {
      sprintf(buf, "nop");
      return 0;
    }
  if ((opcode>>4) == 0x02800)
    {
      if (opcode & 0xf)
	sprintf(buf, "idle(%d)", opcode & 0xf);
      else
	sprintf(buf, "idle");
      return 0;
    }

  /* full codes... */
  if (opcode & BIT23) /* 1 */
    {
      if (opcode & BIT22) /* 11 */
	{
	  if (opcode & BIT21) /* 111 */
	    {
	      sprintf(buf, "111");
	    }
	  else /* 110 */
	    {
	      sprintf(buf, "110");
	    }
	}
      else /* 10 */
	{
	  if (opcode & BIT21) /* 101 */
	    {
	      unsigned int dag = (opcode >> 20)&0x1;
	      unsigned int data = (opcode >> 4)&0x1ffff;
	      unsigned int i = (opcode >> 2)&0x3;
	      unsigned int m = opcode&0x3;
	      sprintf(buf, "dm(i%d,m%d) = 0x%04x", (!dag)?i:i+4, (!dag)?m:m+4, data);
	    }
	  else /* 100 */
	    {
	      if (opcode & BIT20) /* 1001 */
		{
		  unsigned int addr = (opcode >> 4) & 0x3fff;
		  unsigned int rgp = (opcode >> 18) & 0x3;
		  unsigned int reg = opcode & 0xf;
		  sprintf(buf, "dm(0x%04x) = %s", addr, regs[reg][rgp]);
		}
	      else /* BIT20 off */
		{
		  unsigned int addr = (opcode >> 4) & 0x3fff;
                  unsigned int rgp = (opcode >> 18) & 0x3;
                  unsigned int reg = opcode & 0xf;
                  sprintf(buf, "%s = dm(0x%04x)", regs[reg][rgp], addr);
		}
	    }
	}
    }
  else /* BIT23 off */
    {
      if (opcode & BIT22) /* 01 */
	{
	  if (opcode & BIT21) /* 011 */
	    {
	      /* dm(i, m) = dreg */
	      unsigned int dag = (opcode & BIT20)?1:0;
	      if ((opcode & BIT19) && !(opcode & BIT18))
		{
		  unsigned int amf = (opcode >> 13) & 0x1f;
                  unsigned int i = (opcode >> 2) & 0x3;
                  unsigned int m = opcode & 0x3;
		  unsigned int dreg = (opcode >> 4) & 0xf;
		  sprintf(buf, "dm(i%d, m%d) = %s %s", dag?i+4:i, dag?m+4:m, regs[dreg][0], amf?"(AMF unhandled)":"");
		}
	      else /* no amf */
		{
		  unsigned int data = (opcode >> 4) & 0xff;
		  unsigned int i = (opcode >> 1) & 0x1;
		  unsigned int m = opcode & 0x1;
		  sprintf(buf, "dm(i%d, m%d) = 0x%02x", i, m, data); 
		}
	    }
	  else /* 010 */
	    {
	      if (opcode & BIT20) /* 0101 */
		{
		  if (opcode & BIT19) /* 0101 1 */
		    {
		      if (opcode & BIT18) /* 0101 11 */
			{
			  sprintf(buf, "0101 11");
			}
		      else /* 0101 10 */
			{
#if 0
			  unsigned int amf = (opcode >> 13) & 0x1f;
			  unsigned int dreg = (opcode >> 4) & 0xf;
			  unsigned int i = (opcode >> 1) & 0x1;
			  unsigned int m = opcode & 0x1;
#endif
			  sprintf(buf, "0101 10");
 			}
		    }
		  else /* 01010 */
		    {
		      sprintf(buf, "01010");
		    }
		}
	      else /* 0100 */
		{
		  unsigned int addr = (opcode >> 4) & 0x3fff;
		  unsigned int dreg = opcode & 0xf;
		  sprintf(buf, "%s = 0x%04x", regs[dreg][0], addr);
		}
	    }
	}
      else  /* 00 */
	{
	  if (opcode & BIT21) /* 001 */
	    {
	      if (opcode & BIT20) /* 0011 */
		{
		  unsigned int addr = (opcode >> 4) & 0x3fff;
		  unsigned int rgp = (opcode >> 18) & 0x3;
		  unsigned int reg = opcode & 0xf;
		  sprintf(buf, "%s = 0x%04x", regs[reg][rgp], addr);
		}
	      else /* 0010 */
		{
		  if (opcode & BIT19) /* 0010 1 */
		    {
		      unsigned int z = (opcode >> 18) & 0x1;
		      unsigned int amf = (opcode >> 13) & 0x1f;
		      unsigned int yop = (opcode >> 11) & 0x3;
		      unsigned int xop = (opcode >> 8) & 0x3;
		      unsigned int destdreg = (opcode >> 4) & 0xf;
		      unsigned int srcdreg = opcode & 0xf;
		      sprintf(buf, "%s = ", z?"af":"ar");
		      switch(amf)
			{
			case 0x10:
			  sprintf(buf+strlen(buf), "%s", yop?yops[yop]:""); break;
			case 0x11:
			  if (yop)
			    sprintf(buf+strlen(buf), "%s + 1", yops[yop]);
			  else
			    sprintf(buf+strlen(buf), "pass 1");
			  break;
			case 0x12:
			  sprintf(buf+strlen(buf), "%s + %s + c", xops[xop], yops[yop]);
			  break;
			case 0x13:
			  if (yop)
			    sprintf(buf+strlen(buf), "%s + %s", xops[xop], xops[yop]);
			  else
			    sprintf(buf+strlen(buf), "%s", xops[xop]);
			  break;
			default:
			  fprintf(stderr, "decode: unhandled amf %02x\n", amf);
			}
			      
		    }
		  else /* 0010 0 */
		    {
		      unsigned int z = (opcode >> 18) & 0x1;
		      unsigned int amf = (opcode >> 13) & 0x1f;
		      unsigned int yop = (opcode >> 11) & 0x3;
		      unsigned int xop = (opcode >> 8) & 0x3;
		      unsigned int cond = opcode & 0xf;
		      sprintf(buf, "%s = ", z?"af":"ar");
		      switch(amf)
			{
			case 0x10:
			  sprintf(buf+strlen(buf), "%s", yop?yops[yop]:""); break;
			case 0x11:
			  sprintf(buf+strlen(buf), "%s + 1", yops[yop]);
			  break;
			case 0x12:
			  sprintf(buf+strlen(buf), "%s + %s + c", xops[xop], yops[yop]);
			  break;
			case 0x13:
			    sprintf(buf+strlen(buf), "%s + %s", xops[xop], yops[yop]);
			  break;
			case 0x1c:
			  sprintf(buf+strlen(buf), "a%s and 0x%04x", xops[xop], getconstant((opcode>>11)&0x3, (opcode>>6)&0x3, (opcode>>4)&0x3));
			  break;
			default:
			  fprintf(stderr, "decode: unhandled amf\n");
			}
			      
		    }
		}
	    }
	  else /* 000 */
	    {
	      if (opcode & BIT20) /* 0001 */
		{
		  if (opcode & BIT19) /* 00011 */
		    {
		      if (opcode & BIT18) /* 000111 */
			{			  
			  unsigned int cond = opcode & 0xf;
			  unsigned int addr = (opcode >> 4) & 0x3fff;
			  if (cond < 0xf)
			    sprintf(buf, "if %s call 0x%04x", conds[cond], addr);
			  else
			    sprintf(buf, "call 0x%04x", addr);
			}
		      else /* 000110 */
			{
			  unsigned int cond = opcode & 0xf;
			  unsigned int addr = (opcode >> 4) & 0x3fff;
			  if (cond < 0xf)
			    sprintf(buf, "if %s jump 0x%04x", conds[cond], addr);
			  else
			    sprintf(buf, "jump 0x%04x", addr);
			}
		    }
		  else /* 00010 */
		    {
		      if (opcode & BIT18)
			{
			  unsigned short addr,cond;
			  addr = (opcode>>4) & 0x3fff;
			  cond = opcode & 0xf;
			  sprintf(buf, "do %04x until %s", addr, conds[cond]);
			}
		      else
			sprintf(buf, "000101");
		    }
		}
	      else /* 0000 */
		{
		  if (opcode & BIT19) /* 0000 1 */
		    {
		      if (opcode & BIT18) /* 0000 11 */
			{
			  if (opcode & BIT17) /* 0000 111 */
			    {
			      if (opcode & BIT16) /* 0000 1111 */
				{
				  unsigned int sf = (opcode >> 11) & 0xf;
				  unsigned int xop = (opcode >> 8) & 0x7;
				  unsigned int exp = opcode & 0xff;
				  sprintf(buf, "sr = %s %s %s by 0x%04x %s", 
					  sfs[sf][2][0]==1?"sr or":"",
					  sfs[sf][0],
					  xops[xop],
					  exp,
					  sfs[sf][1]);
				}
			      else /* 0000 1110 */
				{
				  sprintf(buf, "0000 1110");
				}
			    }
			  else /* 0000 110 */
			    {
			      if (opcode & BIT16)
				{
				  sprintf(buf, "%s = %s", 
					  regs[(opcode>>4)&0xf][(opcode>>10)&0x3], regs[opcode&0xf][(opcode>>8)&0x3]);
				}
			      else /* 0000 1100 */
				{
				  buf[0] = '\0';
				  
				  if (((opcode>>2)&03)>=2)
				    {
				      if (((opcode>>2)&0x3)==2)
					strcat(buf, "dis g_mode ");
				      else
					strcat(buf, "ena g_mode ");
				    }

				  if (((opcode>>4)&03)>=2)
				    {
				      if (((opcode>>4)&0x3)==2)
					strcat(buf, "dis sec_reg ");
				      else
					strcat(buf, "ena sec_reg ");
				    }

				  if (((opcode>>6)&03)>=2)
				    {
				      if (((opcode>>6)&0x3)==2)
					strcat(buf, "dis bit_rev ");
				      else
					strcat(buf, "ena bit_rev ");
				    }

				  if (((opcode>>8)&03)>=2)
				    {
				      if (((opcode>>8)&0x3)==2)
					strcat(buf, "dis av_latch ");
				      else
					strcat(buf, "ena av_latch ");
				    }

				  if (((opcode>>0x0a)&03)>=2)
				    {
				      if (((opcode>>0x0a)&0x3)==2)
					strcat(buf, "dis ar_sat ");
				      else
					strcat(buf, "ena ar_sat ");
				    }

				  if (((opcode>>0x0c)&03)>=2)
				    {
				      if (((opcode>>0x0c)&0x3)==2)
					strcat(buf, "dis m_mode ");
				      else
					strcat(buf, "ena m_mode ");
				    }

				  if (((opcode>>0x0e)&03)>=2)
				    {
				      if (((opcode>>0x0e)&0x3)==2)
					strcat(buf, "dis timer ");
				      else
					strcat(buf, "ena timer ");
				    }
				}
				
			    }
			}
		      else /* 0000 10 */
			{
			  if ((opcode >> 4) == 0xa001) /* rti */
			    {
			      unsigned int cond = opcode & 0xf;
			      if (cond < 0xf)
				sprintf(buf, "if %s rti", conds[cond]);
			      else
				sprintf(buf, "rti");
			    }
			  else if ((opcode >> 4) == 0xa000) /* rts */
			    {
                              unsigned int cond = opcode & 0xf;
                              if (cond < 0xf)
                                sprintf(buf, "if %s rts", conds[cond]);
                              else
                                sprintf(buf, "rts");
			    }
			  else if ((opcode >> 4) == 0x09080)
			    {
			      unsigned int dag = (opcode >> 4) & 0x1;
			      unsigned int i = (opcode >> 2) & 0x3;
			      unsigned int m = opcode & 0x3;
			      sprintf(buf, "modify (i%d, m%d)", dag?i+4:i, dag?m+4:m);
			    }
			  else if ((opcode>>8) == 0xb00)
			    {
			      sprintf(buf, "%s%s %s (i%d)",
				      (opcode&0xf)?"if ":"",
				      (opcode&0xf)?conds[(opcode&0xf)]:"",
				      ((opcode>>4)&0x1)?"call":"jump",
				      (opcode>>6)&0x8);
			    }
			}
		    }
		  else /* 0000 0 */
		    {
		      if (opcode & BIT18) /* 0000 01 */
			{
			  if ((opcode >> 5) == 0x2000)
			    {
			      /* stack control */
			      sprintf(buf, "[stack control]");
			    }

			  if (opcode & BIT17) /* 0000 011 */
			    {
			      if (opcode & BIT16) /* 0000 0111 */
				{
				  /* DIVQ */
				  unsigned int xop = (opcode >> 8) & 0x7;
				  sprintf(buf, "divq %s", xops[xop]);
				}
			      else /* 0000 0110 */
				{
				  /* DIVS */
				  unsigned int yop = (opcode >> 11) & 0x3;
				  unsigned int xop = (opcode >> 8) & 0x7;
				  sprintf(buf, "divs %s, %s", yops[yop], xops[xop]);
				}
			    }
			  else /* 0000 010 */
			    {
			      sprintf(buf, "0000 010");
			    }
			}
		      else /* 000000 */
			{
			  if (opcode & BIT17) /* 0000 001 */
			    {
			      if (opcode & BIT16) /* 0000 0011 */
				{
				  /* call/jump on flag in */
				  unsigned int addr = (opcode >> 4) & 0xfff;
				  unsigned int addrmsb = (opcode >> 2) & 0x3;
				  unsigned int fic = (opcode >> 1) & 0x1;
				  unsigned int s = (opcode & 0x1);
				  sprintf(buf, "if %sfi %s 0x%04x",
					  fic?"":"not ",
					  s?"call":"jump",
					  (addrmsb<<12)+addr);
		     
				}
			      else /* 0000 0010 */
				{
				  /* modify flag out */
				  unsigned int fl[4] = {
				    (opcode >> 10) & 0x3,
				    (opcode >> 8) & 0x3,
				    (opcode >> 6) & 0x3,
				    (opcode >> 4) & 0x3,
				  };
				  unsigned int cond = opcode & 0xf;
				  int cur = 0, i = 0;

				  if (cond < 0xf)
				    {
				      sprintf(buf, "if %s ", conds[cond]);
				      i += 6;
				    }
				  for (cur=0;cur<3;cur++)
				    {
				      if (fl[cur])
					{
					  sprintf(&(buf[i]), "%s fl%d ",
						  flagops[fl[cur]],
						  cur);
					  i += strlen(flagops[fl[cur]]) + 5;
					}
				    }
				  if (fl[3])
				    {
				      sprintf(&(buf[i]), "%s flag_out", 
					      flagops[fl[3]]);
				    }
				  
				}
			    }
			  else /* 0000 000 */
			    {
			      if (opcode & BIT16) /* 0000 0001 */
				{
				  unsigned int addr = (opcode >> 4) & 0x7ff;
				  unsigned int dreg = opcode & 0xf;
				  unsigned int dir = opcode & BIT15;
				  if (dir)
				    sprintf(buf, "io(%d) = %s", addr, regs[dreg][0]);
				  else
				    sprintf(buf, "%s = io(%d)", regs[dreg][0], addr);
				}
			      else /* 0000 0000 */
				{
				  sprintf(buf, "0000 0000");
				}
			    }
			}
		    }
		}
	    }
	}
    }
  return 0;
}
